/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2011
	Copyright Stichting C.A. Muller Radioastronomiestation, 2011

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <controller/controller_block.h>

#include "block_beckhoff_el2502.h"
#include "ec.h"
#include "esc.h"

#define EL2502_COE_RXPDO_ASSIGN		0x1c12
#define EL2502_COE_RXPDO_ASSIGN_O1_PWM		0x1600	/* only PWM */
#define EL2502_COE_RXPDO_ASSIGN_O2_PWM		0x1601	/* only PWM */
#define EL2502_COE_RXPDO_ASSIGN_O1_PWM_PER	0x1602	/* PWM and period */
#define EL2502_COE_RXPDO_ASSIGN_O2_PWM_PER	0x1603	/* PWM and period */

#define EL2502_PERIOD_20	0.05
#define EL2502_PERIOD_20000	0.00005

struct controller_block_private {
	float	*output1;
	float	*period1;
	float	*output2;
	float	*period2;
	
	unsigned char *tx_buffer;
};

static void calculate(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	float output1, output2, period1, period2;
	union {
		uint8_t  output_c[8];
		uint16_t output_u[4];
	} output;
	
	output1 = fmax(1.0, fabs(*priv->output1)) * 65535.0;
	output2 = fmax(1.0, fabs(*priv->output2)) * 65535.0;

	period1 = fmax(EL2502_PERIOD_20000, fmin(EL2502_PERIOD_20, fabs(*priv->period1))) * 1000000.0;
	period2 = fmax(EL2502_PERIOD_20000, fmin(EL2502_PERIOD_20, fabs(*priv->period2))) * 1000000.0;

	output.output_u[0] = (uint16_t)output1;
	output.output_u[1] = (uint16_t)period1;
	output.output_u[2] = (uint16_t)output2;
	output.output_u[3] = (uint16_t)period2;

	memcpy(priv->tx_buffer, output.output_c, sizeof(output.output_c));
	
	ec_tx_pdo(controller_samplenr);
}


struct beckhoff_el2502_type {
	char *name;
	uint32_t productcode;
};

static struct beckhoff_el2502_type devlist[] = {
	{ "EL2502", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2502 },
};

struct controller_block * block_beckhoff_el2502_create(char *name, va_list ap)
{
	struct controller_block *block;
	struct ec_dgram_addr addr;
	struct esc_syncmanager_info pdo_tx, pdo_rx;
	int slaves;
	int i;
	ssize_t r;
	bool found;
	int dev = -1;
	char *eltype;
	int devno;
	struct esc_mailbox *mb;
	struct esc_syncmanager_info mb_rd, mb_wr;
	struct canopen_dev *canopen_dev;
	uint16_t pdo_assign;
	struct timespec timeout = { 1, 0 };
		
	eltype = va_arg(ap, char *);
	devno = va_arg(ap, int);
	
	for (i = 0; i < sizeof(devlist)/sizeof(struct beckhoff_el2502_type); i++) {
		if (!strcmp(eltype, devlist[i].name)) {
			dev = i;
			break;
		}
	}
	if (dev < 0)
		return NULL;
	
	slaves = ec_slave_count();
	found = false;
	for (i = 0; i < slaves; i++) {
		struct ec_dgram_addr devaddr;
		uint32_t vendor, product;
		
		ec_addr_set_auto_inc_nr(&devaddr, i);
		vendor = esc_esi_vendorid_get(&devaddr);
		product = esc_esi_productcode_get(&devaddr);
		if (vendor == ESC_ESI_VENDORID_BECKHOFF &&
		    product == devlist[dev].productcode) {
			if (!devno || devno == i) {
				found = true;
				addr = devaddr;
			}
		}
	}
	if (!found) {
		printf("Device not found\n");
		return NULL;
	}
	
	block = malloc(sizeof(struct controller_block));
	if (!block)
		return NULL;

	block->private = malloc(sizeof(struct controller_block_private));
	if (!block->private)
		return NULL;

	block->type = "beckhoff_el2502";
	block->name = strdup(name);

	esc_init(&addr);

	pdo_tx.start = 0x1100;
	pdo_tx.len = 0x8;
	pdo_tx.sm = 2;
	
	pdo_rx.start = 0x1180;
	pdo_rx.len = 0;
	pdo_rx.sm = 3;

	esc_pdo_tx_set(&addr, &pdo_tx, &block->private->tx_buffer, true);
	esc_pdo_rx_set(&addr, &pdo_rx, NULL);

	if (esc_al_state_set(&addr, ESC_AL_STATE_INIT, &timeout) < 0)
		printf("Could not go to state init\n");

	/* mailbox write (master->slave) @0x1000 */
	mb_wr.start = 0x1000;
	mb_wr.len = 128;
	mb_wr.sm = 0;
	
	/* mailbox read (slave->master) @0x1080 */
	mb_rd.start = 0x1080;
	mb_rd.len = 128;
	mb_rd.sm = 1;
	
	mb = esc_mailbox_create(&addr, &mb_rd, &mb_wr);
	canopen_dev = esc_coe_create(mb);

	/* Set the rxpdo such that we can set both pwm and period values */

	r = canopen_read_param(canopen_dev, 
	    EL2502_COE_RXPDO_ASSIGN, 1, &pdo_assign, sizeof(pdo_assign));
	if (r != sizeof(pdo_assign)) {
		printf("Expected %zd from read, but got %zd!\n",
		    sizeof(pdo_assign), r);
	}
	printf("RxPDO assign for output 1: 0x%04x\n", le16toh(pdo_assign));
	r = canopen_read_param(canopen_dev, 
	    EL2502_COE_RXPDO_ASSIGN, 2, &pdo_assign, sizeof(pdo_assign));
	if (r != sizeof(pdo_assign)) {
		printf("Expected %zd from read, but got %zd!\n",
		    sizeof(pdo_assign), r);
	}
	printf("RxPDO assign for output 2: 0x%04x\n", le16toh(pdo_assign));

	pdo_assign = htole16(EL2502_COE_RXPDO_ASSIGN_O1_PWM_PER);
	r = canopen_write_param(canopen_dev,
	    EL2502_COE_RXPDO_ASSIGN, 1, &pdo_assign, sizeof(pdo_assign));
	if (r != sizeof(pdo_assign)) {
		printf("Expected %zd from write, but got %zd!\n",
		    sizeof(pdo_assign), r);
	}
	pdo_assign = htole16(EL2502_COE_RXPDO_ASSIGN_O2_PWM_PER);
	r = canopen_write_param(canopen_dev,
	    EL2502_COE_RXPDO_ASSIGN, 2, &pdo_assign, sizeof(pdo_assign));
	if (r != sizeof(pdo_assign)) {
		printf("Expected %zd from write, but got %zd!\n",
		    sizeof(pdo_assign), r);
	}

	r = canopen_read_param(canopen_dev, 
	    EL2502_COE_RXPDO_ASSIGN, 1, &pdo_assign, sizeof(pdo_assign));
	if (r != sizeof(pdo_assign)) {
		printf("Expected %zd from read, but got %zd!\n",
		    sizeof(pdo_assign), r);
	}
	printf("RxPDO assign for output 1: 0x%04x\n", le16toh(pdo_assign));
	r = canopen_read_param(canopen_dev, 
	    EL2502_COE_RXPDO_ASSIGN, 2, &pdo_assign, sizeof(pdo_assign));
	if (r != sizeof(pdo_assign)) {
		printf("Expected %zd from read, but got %zd!\n",
		    sizeof(pdo_assign), r);
	}
	printf("RxPDO assign for output 2: 0x%04x\n", le16toh(pdo_assign));

	if (esc_al_state_set(&addr, ESC_AL_STATE_PRE_OPERATIONAL, &timeout) < 0)
		printf("Could not go to state pre operational\n");
	if (esc_al_state_set(&addr, ESC_AL_STATE_SAFE_OPERATIONAL, &timeout) < 0)
		printf("Could not go to state safe operational\n");
	if (esc_al_state_set(&addr, ESC_AL_STATE_OPERATIONAL, &timeout) < 0)
		printf("Could not go to state safe operational\n");
	printf("state: %d\n", esc_al_state_get(&addr));
	printf("statuscode: %d\n", esc_al_status_code_get(&addr));

	block->outputs = 0;
	block->output = NULL;
	
	/* Outputs on the hardware are inputs on this software block */
	block->inputs = 4;
	block->input = malloc(sizeof(struct controller_block_interm) * 4);
	if (!block->input)
		return NULL;
	block->input[0].name = "output1";
	block->input[0].type = CONTROLLER_BLOCK_TERM_FLOAT;
	block->input[0].value.f = &block->private->output1;
	block->input[0].ghostof = NULL;
	block->input[1].name = "period1";
	block->input[1].type = CONTROLLER_BLOCK_TERM_FLOAT;
	block->input[1].value.f = &block->private->period1;
	block->input[1].ghostof = NULL;
	block->input[2].name = "output2";
	block->input[2].type = CONTROLLER_BLOCK_TERM_FLOAT;
	block->input[2].value.f = &block->private->output2;
	block->input[2].ghostof = NULL;
	block->input[3].name = "period2";
	block->input[3].type = CONTROLLER_BLOCK_TERM_FLOAT;
	block->input[3].value.f = &block->private->period2;
	block->input[3].ghostof = NULL;
	
	block->calculate = calculate;

	block->params = 0;
	block->param = NULL;
	block->param_get = NULL;
	block->param_set = NULL;

	controller_block_add(block);
	return block;
}
