/*
	Copyright Daan Vreeken <Daan@Vitsch.nl> - Vitsch Electronics, 2012
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2009
	Copyright Stichting C.A. Muller Radioastronomiestation, 2009

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include <controller/controller_block.h>

#include "block_beckhoff_el3xxx.h"
#include "ec.h"
#include "esc.h"
#include "log.h"


struct controller_block_private {
	float			input[8];
	int			check_error_mask;
	int			debug;
	
	bool			have_errors;
	
	float			gain;
	float			offset;
	
	unsigned char		*rx_buffer;
	int		 	inputs;
};

struct el3_24bit_ch {
	uint8_t			status;
	uint16_t		value;
} __packed;

struct el3_24bit_2ch_pdo {
	struct el3_24bit_ch	ch[2];
} __packed;


static void calculate_2s(struct controller_block *block)
{
	struct controller_block_private	*priv = block->private;
	struct el3_24bit_2ch_pdo	*rxdata;
	int				ch_idx;
	int16_t				value;
	

	rxdata = (struct el3_24bit_2ch_pdo *)priv->rx_buffer;
	ec_rx_pdo(controller_samplenr);

	for (ch_idx = 0; ch_idx < 2; ch_idx++) {
		value = (int16_t)le16toh(rxdata->ch[ch_idx].value);
		priv->input[ch_idx] = (float)value * priv->gain +
		    priv->offset;
	}

	log_send(LOG_T_DEBUG, "el3xxx(%s): _s2: TEST ME! %f %f",
	    block->name, priv->input[0], priv->input[1]);
}

static void calculate_2u(struct controller_block *block)
{
	struct controller_block_private	*priv = block->private;
	struct el3_24bit_2ch_pdo	*rxdata;
	int				ch_idx;
	uint16_t			value;
	
	
	rxdata = (struct el3_24bit_2ch_pdo *)priv->rx_buffer;
	ec_rx_pdo(controller_samplenr);

	for (ch_idx = 0; ch_idx < 2; ch_idx++) {
		value = le16toh(rxdata->ch[ch_idx].value);
		priv->input[ch_idx] = (float)value * priv->gain +
		    priv->offset;
	}

	log_send(LOG_T_DEBUG, "el3xxx(%s): _2u: TEST ME! %f %f",
	    block->name, priv->input[0], priv->input[1]);
}


struct el3_32bit_ch {
	uint16_t		status;
	uint16_t		value;
} __packed;

struct el3_32bit_4ch_pdo {
	struct el3_32bit_ch	ch[4];
} __packed;

#define FLAG16_UNDERRANGE	(1 << 0)
#define FLAG16_OVERRANGE	(1 << 1)
#define FLAG16_LIMIT1		(0x03 << 2)
#define FLAG16_LIMIT2		(0x03 << 5)
#define FLAG16_ERROR		(1 << 7)
#define FLAG16_PDO_STATE	(1 << 14)
#define FLAG16_PDO_TOGGLE	(1 << 15)
#define FLAG16_MASK		(FLAG16_UNDERRANGE | FLAG16_OVERRANGE |	\
    FLAG16_LIMIT1 | FLAG16_LIMIT2 | FLAG16_ERROR)


static void print_flags16(char *name, int ch_nr, uint16_t state)
{
	printf("%s: ch%d:%s%s%s%s%s\n", name, ch_nr,
	    (state & FLAG16_UNDERRANGE) ? " underrange" : "",
	    (state & FLAG16_OVERRANGE) ? " overrange" : "",
	    (state & FLAG16_LIMIT1) ? " limit1" : "",
	    (state & FLAG16_LIMIT2) ? " limit2" : "",
	    (state & FLAG16_ERROR) ? " error" : "");
}

static void calculate_4u_32bit(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	int				ch_idx;
	uint16_t			value, state;
	struct el3_32bit_4ch_pdo	*rxdata;
	bool				have_errors;
	
	
	rxdata = (struct el3_32bit_4ch_pdo *)priv->rx_buffer;
	ec_rx_pdo(controller_samplenr);
	
	have_errors = false;
	for (ch_idx = 0; ch_idx < 4; ch_idx++) {
		value = le16toh(rxdata->ch[ch_idx].value);
		priv->input[ch_idx] = (float)value * priv->gain +
		    priv->offset;

		state = rxdata->ch[ch_idx].status;
		if ((priv->check_error_mask & (1 << ch_idx)) &&
		    (state & FLAG16_MASK)) {
			have_errors = true;
			print_flags16(block->name, ch_idx + 1, state);
		}
	}
	priv->have_errors = have_errors;
	
	if (priv->debug)
		printf("%s: 4u_32bit: %0.3f %0.3f %0.3f %0.3f\n", block->name,
		    priv->input[0], priv->input[1], priv->input[2],
		    priv->input[3]);
}


static void calculate_4s_32bit(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	int				ch_idx;
	uint16_t			value, state;
	struct el3_32bit_4ch_pdo	*rxdata;
	bool				have_errors;
	

	rxdata = (struct el3_32bit_4ch_pdo *)priv->rx_buffer;
	ec_rx_pdo(controller_samplenr);

	have_errors = false;
	for (ch_idx = 0; ch_idx < 4; ch_idx++) {
		value = le16toh(rxdata->ch[ch_idx].value);
		priv->input[ch_idx] = (float)value * priv->gain +
		    priv->offset;

		state = le16toh(rxdata->ch[ch_idx].status);
		if ((priv->check_error_mask & (1 << ch_idx)) &&
		    (state & FLAG16_MASK)) {
			have_errors = true;
			print_flags16(block->name, ch_idx + 1, state);
		}
	}
	priv->have_errors = have_errors;

	if (priv->debug)
		printf("%s: 4s_32bit: %0.2f %0.2f %0.2f %0.2f\n", block->name,
		    priv->input[0], priv->input[1], priv->input[2],
		    priv->input[3]);
}

struct beckhoff_el3xxx_type {
	char *name;
	uint32_t productcode;
	int inputs;
	float gain;
	float offset;
	bool sign;
	int pdo_bits_per_ch;
	int driver_verified;
	struct controller_block_outterm_list *outterms;
};


static struct controller_block_outterm_list outterms_1[] = {
	{ "input1", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input[0])  },
	{ "error",  CONTROLLER_BLOCK_TERM_BOOL,  offsetof(struct controller_block_private, have_errors) },
	{ NULL },
};

static struct controller_block_outterm_list outterms_2[] = {
	{ "input1", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input[0])  },
	{ "input2", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input[1])  },
	{ "error",  CONTROLLER_BLOCK_TERM_BOOL,  offsetof(struct controller_block_private, have_errors) },
	{ NULL },
};

static struct controller_block_outterm_list outterms_4[] = {
	{ "input1", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input[0])  },
	{ "input2", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input[1])  },
	{ "input3", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input[2])  },
	{ "input4", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input[3])  },
	{ "error",  CONTROLLER_BLOCK_TERM_BOOL,  offsetof(struct controller_block_private, have_errors) },
	{ NULL },
};

static struct controller_block_outterm_list outterms_8[] = {
	{ "input1", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input[0])  },
	{ "input2", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input[1])  },
	{ "input3", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input[2])  },
	{ "input4", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input[3])  },
	{ "input5", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input[4])  },
	{ "input6", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input[5])  },
	{ "input7", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input[6])  },
	{ "input8", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input[7])  },
	{ "error",  CONTROLLER_BLOCK_TERM_BOOL,  offsetof(struct controller_block_private, have_errors) },
	{ NULL },
};
static struct beckhoff_el3xxx_type devlist[] = {
	/*
	 * name, productcode, inputs, gain, offset, sign, pdo_bits_per_ch,
	 * driver_verified
	 */
	{ "EL3001", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3001, 1, 10.0/32767, 0.0, true, 24, 0, outterms_1 },
	{ "EL3002", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3002, 2, 10.0/32767, 0.0, true, 24, 0, outterms_2 },
	{ "EL3004", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3004, 4, 10.0/32767, 0.0, true, 24, 0, outterms_4 },
	{ "EL3008", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3008, 8, 10.0/32767, 0.0, true, 24, 0, outterms_8 },
	{ "EL3041", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3041, 1, 0.02/65536, 0.0, false, 24, 0, outterms_1 },
	{ "EL3042", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3042, 2, 0.02/65536, 0.0, false, 24, 0, outterms_2 },
	{ "EL3044", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3044, 4, 0.02/65536, 0.0, false, 24, 0, outterms_4 },
	{ "EL3048", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3048, 8, 0.02/65536, 0.0, false, 24, 0, outterms_8 },
	{ "EL3051", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3051, 1, 0.02/65536, 0.004, false, 24, 0, outterms_1 },
	{ "EL3052", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3052, 2, 0.02/65536, 0.004, false, 24, 0, outterms_2 },
	{ "EL3054", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3054, 4, 0.02/65536, 0.004, false, 24, 0, outterms_4 },
	{ "EL3058", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3058, 8, 0.02/65536, 0.004, false, 24, 0, outterms_8 },
	{ "EL3061", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3061, 1, 10.0/65536, 0.0, false, 24, 0, outterms_1 },
	{ "EL3062", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3062, 2, 10.0/65536, 0.0, false, 24, 0, outterms_2 },
	{ "EL3064", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3064, 4, 10.0/65536, 0.0, false, 24, 0, outterms_4 },
	{ "EL3068", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3068, 8, 10.0/65536, 0.0, false, 24, 0, outterms_8 },

	{ "EL3102", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3102, 2, 10.0 / 32767, 0.0, true, 24, 0, outterms_2 },
	/* current inputs 4..20mA */
	{ "EL3124", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3124, 4, 0.016 / 32767, 0.004, false, 32, 1, outterms_4 },
	/* temperature sensors (PT100/1000 etc) */
	{ "EL3204", ESC_ESI_PRODUCTCODE_BECKHOFF_EL3204, 4, 0.1, 0.0, true, 32, 1, outterms_4 },
};

#define DEVLIST_LEN	(sizeof(devlist) / sizeof(struct beckhoff_el3xxx_type))


static struct controller_block_param_list params[] = {
	/* KEEP IN THIS ORDER! (we use the index..) */
	{ "check_input1", false },
	{ "check_input2", false },
	{ "check_input3", false },
	{ "check_input4", false },
	{ "check_input5", false },
	{ "check_input6", false },
	{ "check_input7", false },
	{ "check_input8", false },
	{ "debug", false },
	{ NULL },
};

#define NR_OF_PARAMS	(sizeof(params) / sizeof(struct controller_block_param_list))

static void param_get(struct controller_block *block, int param, void *val)
{
	switch (param) {
	case 8:
		*(bool *)val = block->private->debug;
		return;
	}

	// must be check_input[1..8] when we reach this point
	if (block->private->check_error_mask & (1 << param))
		*(bool *)val = true;
	else
		*(bool *)val = false;
}

static void param_set(struct controller_block *block, int param, va_list val)
{
	int		new_val;
	
	new_val = va_arg(val, int);

	switch (param) {
	case 8:
		block->private->debug = new_val;
		return;
	}
	
	// must be check_input[1..8] when we reach this point
	if (new_val)
		// enable error checking on this input
		block->private->check_error_mask |= (1 << param);
	else
		// disable error checking on this input
		block->private->check_error_mask &= ~(1 << param);
}


struct controller_block * block_beckhoff_el3xxx_create(char *name, va_list ap)
{
	struct controller_block		*block;
	struct ec_dgram_addr		addr;
	struct esc_syncmanager_info	pdo_rx, pdo_tx;
	int				slaves;
	int				i;
	bool				found;
	struct beckhoff_el3xxx_type	*dev;
	char				*eltype;
	int				devno;
	struct esc_mailbox		*mb;
	struct esc_syncmanager_info	mb_rd, mb_wr;
	struct timespec			timeout = { 1, 0 };


	eltype = va_arg(ap, char *);
	devno = va_arg(ap, int);
	
	dev = NULL;
	for (i = 0; i < DEVLIST_LEN; i++) {
		if (!strcmp(eltype, devlist[i].name)) {
			dev = &devlist[i];
			break;
		}
	}
	if (dev == NULL)
		return NULL;
	
	slaves = ec_slave_count();
	found = false;
	for (i = 0; i < slaves; i++) {
		struct ec_dgram_addr devaddr;
		uint32_t vendor, product;
		
		ec_addr_set_auto_inc_nr(&devaddr, i);
		vendor = esc_esi_vendorid_get(&devaddr);
		product = esc_esi_productcode_get(&devaddr);
		if (vendor == ESC_ESI_VENDORID_BECKHOFF &&
		    product == dev->productcode) {
			if (!devno || devno == i) {
				found = true;
				addr = devaddr;
			}
		}
	}
	if (!found) {
		log_send(LOG_T_ERROR, "%s: Device not found\n", name);
		return NULL;
	}
	if (! dev->driver_verified) {
		log_send(LOG_T_ERROR, 
		    "el3xxx: This driver hasn't (yet) been verified to "
		    "with the %s. Be very careful!", dev->name);
	}
	
	block = controller_block_alloc("beckhoff_el3xxx", name, 
	    sizeof(struct controller_block_private));
	if (!block)
		return NULL;

	esc_init(&addr);

	pdo_rx.start = 0x1180;
	pdo_rx.len = dev->pdo_bits_per_ch * dev->inputs / 8;
	pdo_rx.sm = 3;
	
	pdo_tx.start = 0x1100;
	pdo_tx.len = 0;
	pdo_tx.sm = 2;

	esc_pdo_tx_set(&addr, &pdo_tx, NULL, false);
	esc_pdo_rx_set(&addr, &pdo_rx, &block->private->rx_buffer);

	if (esc_al_state_set(&addr, ESC_AL_STATE_INIT, &timeout) < 0)
		printf("Could not go to state init\n");

	/* mailbox read (slave->master) @0x1080 */
	mb_rd.start = 0x1080;
	mb_rd.len = 128;
	mb_rd.sm = 1;
	
	/* mailbox write (master->slave) @0x1000 */
	mb_wr.start = 0x1000;
	mb_wr.len = 128;
	mb_wr.sm = 0;
	
	mb = esc_mailbox_create(&addr, &mb_rd, &mb_wr);
	esc_coe_create(mb);

	if (esc_al_state_set(&addr, ESC_AL_STATE_PRE_OPERATIONAL, &timeout) < 0)
		printf("Could not go to state pre operational\n");
	if (esc_al_state_set(&addr, ESC_AL_STATE_SAFE_OPERATIONAL, &timeout) < 0)
		printf("Could not go to state safe operational\n");
	if (esc_al_state_set(&addr, ESC_AL_STATE_OPERATIONAL, &timeout) < 0)
		printf("Could not go to state operational\n");


	/* Inputs on the hardware are outputs on this software block */
	if (controller_block_outterm_list_init(block, dev->outterms))
		return NULL;

	block->calculate = NULL;
	
	switch (dev->inputs) {
	case 2:
		if (dev->pdo_bits_per_ch == 16) {
			if (dev->sign)
				block->calculate = calculate_2s;
			else
				block->calculate = calculate_2u;
		}
		break;
	case 4:
		if (dev->pdo_bits_per_ch == 32) {
			if (dev->sign == false)
				block->calculate = calculate_4u_32bit;
			else
				block->calculate = calculate_4s_32bit;
		}
		break;
	default:
		log_send(LOG_T_ERROR, "el3xxx: implement me first! (%d ch)",
		    dev->inputs);
	}
	
	if (block->calculate == NULL)
		log_send(LOG_T_ERROR, 
		    "el3xxx: EEK! implement me first! (calc for inp=%d, "
		    "sign=%s, pdo bits/ch=%d", dev->inputs,
		    dev->sign ? "true" : "false", dev->pdo_bits_per_ch);
	
	block->private->gain = dev->gain;
	block->private->offset = dev->offset;
	
	block->private->check_error_mask = 0;
	block->private->debug = 0;
	
	block->params = NR_OF_PARAMS;
	block->param = params;
	block->param_get = param_get;
	block->param_set = param_set;
	
	controller_block_add(block);
	return block;
}
