/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2009
	Copyright Stichting C.A. Muller Radioastronomiestation, 2009

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include <controller/controller_block.h>

#include "ec.h"
#include "esc.h"
#include "log.h"

struct controller_block_private {
	float	 converted_value;
	uint32_t value;
	
	double 	gain;
	double 	offset;
	
	unsigned char *rx_buffer;
};

static void calculate(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	struct {
		uint8_t	 status;
		uint32_t value;
	} __attribute__((packed)) *rxdata;
	uint32_t value;
	
	rxdata = (void*)priv->rx_buffer;
	
	ec_rx_pdo(controller_samplenr);

	value = le32toh(rxdata->value);
	priv->value = value;
	priv->converted_value = (value + priv->gain) / priv->offset;
}


struct controller_block * block_beckhoff_el5001_create(char *name, va_list ap)
{
	struct controller_block *block;
	struct ec_dgram_addr addr;
	struct esc_syncmanager_info pdo_rx, pdo_tx;
	int slaves;
	int i;
	bool found;
	int devno;
	struct timespec timeout = { 1, 0 };
		
	struct esc_mailbox *mb;
	struct esc_syncmanager_info mb_rd, mb_wr;

	devno = va_arg(ap, int);
	
	slaves = ec_slave_count();
	found = false;
	for (i = 0; i < slaves; i++) {
		struct ec_dgram_addr devaddr;
		uint32_t vendor, product;
		
		ec_addr_set_auto_inc_nr(&devaddr, i);
		vendor = esc_esi_vendorid_get(&devaddr);
		product = esc_esi_productcode_get(&devaddr);
		if (vendor == ESC_ESI_VENDORID_BECKHOFF &&
		    product == ESC_ESI_PRODUCTCODE_BECKHOFF_EL5001) {
			if (!devno || devno == i) {
				found = true;
				addr = devaddr;
			}
		}
	}
	if (!found) {
		log_send(LOG_T_ERROR, "%s: Device not found\n", name);
		return NULL;
	}
	
	block = controller_block_alloc("beckhoff_el5001", name,
	    sizeof(struct controller_block_private));
	if (!block)
		return NULL;

	esc_init(&addr);

	pdo_rx.start = 0x1100;
	pdo_rx.len = 5;
	pdo_rx.sm = 3;
	
	pdo_tx.start = 0x1000;
	pdo_tx.len = 0;
	pdo_tx.sm = 2;

	esc_pdo_tx_set(&addr, &pdo_tx, NULL, true);
	esc_pdo_rx_set(&addr, &pdo_rx, &block->private->rx_buffer);

	if (esc_al_state_set(&addr, ESC_AL_STATE_INIT, &timeout) < 0)
		printf("Could not go to state init\n");

	/* mailbox read (slave->master) @0x18f6 */
	mb_rd.start = 0x18f6;
	mb_rd.len = 246;
	mb_rd.sm = 1;
	
	/* mailbox write (master->slave) @0x1800 */
	mb_wr.start = 0x1800;
	mb_wr.len = 246;
	mb_wr.sm = 0;
	
	mb = esc_mailbox_create(&addr, &mb_rd, &mb_wr);
	esc_coe_create(mb);

	if (esc_al_state_set(&addr, ESC_AL_STATE_PRE_OPERATIONAL, &timeout) < 0)
		printf("Could not go to state pre operational\n");
	if (esc_al_state_set(&addr, ESC_AL_STATE_SAFE_OPERATIONAL, &timeout) < 0)
		printf("Could not go to state safe operational\n");
	if (esc_al_state_set(&addr, ESC_AL_STATE_OPERATIONAL, &timeout) < 0)
		printf("Could not go to state operational\n");

	block->inputs = 0;
	block->input = NULL;
	
	/* Inputs on the hardware are outputs on this software block */
	block->outputs = 2;
	block->output = malloc(sizeof(struct controller_block_outterm) * 8);
	if (!block->output)
		return NULL;
	block->output[0].name = "converted_value";
	block->output[0].type = CONTROLLER_BLOCK_TERM_FLOAT;
	block->output[0].value.f = &block->private->converted_value;
	block->output[0].source = block;
	block->output[1].name = "value";
	block->output[1].type = CONTROLLER_BLOCK_TERM_UINT32;
	block->output[1].value.v = &block->private->value;
	block->output[1].source = block;
	
	block->calculate = calculate;
	
	block->private->gain = 1.0;
	block->private->offset = 0.0;
	
	block->params = 0;
	block->param = NULL;
	block->param_get = NULL;
	block->param_set = NULL;

	controller_block_add(block);
	return block;
}
