/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007, 2013
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007, 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <string.h>
#include <stdio.h>
#include <pthread.h>

#include "canopen.h"

#include "log.h"

#define CANOPEN_CSS_MASK	0xe0
#define CANOPEN_CSS_INITIATE_DOMAIN_DOWNLOAD	(0x1 << 5)	// 0x20
#define CANOPEN_CSS_INITIATE_DOMAIN_UPLOAD	(0x2 << 5)	// 0x40
#define CANOPEN_CSS_ABORT			(0x4 << 5)	// 0x80
#define CANOPEN_N_MASK		0x0c
#define CANOPEN_N_SHIFT		2
#define CANOPEN_E		0x02
#define CANOPEN_S		0x01

struct timespec canopen_timeout = { 1, 0 };

ssize_t canopen_write_param(struct canopen_dev *dev,
    unsigned index, unsigned subindex,
    void *data, ssize_t bytes)
{
	unsigned char cmd[8];
	ssize_t ret;
	
	memcpy(cmd + 4, data, bytes);
	cmd[0] =
	    CANOPEN_CSS_INITIATE_DOMAIN_DOWNLOAD |
	    ((4-bytes) << CANOPEN_N_SHIFT) |
	    CANOPEN_E | CANOPEN_S;
	cmd[1] = index & 0xff;
	cmd[2] = index >> 8;
	cmd[3] = subindex & 0xff;

	pthread_mutex_lock(&dev->lock);

	ret = dev->send_request(dev, cmd, 8, &canopen_timeout);
	if (ret == 8)
		ret = dev->recv_response(dev, cmd, 8, &canopen_timeout);

	pthread_mutex_unlock(&dev->lock);

	if (ret != 8)
		return -1;
	
	if ((cmd[0] & CANOPEN_CSS_MASK) == CANOPEN_CSS_ABORT)
		return -1;
	
	return bytes;
}

ssize_t canopen_read_param(struct canopen_dev *dev,
    unsigned index, unsigned subindex,
    void *data, ssize_t bytes)
{
	unsigned char cmd[8];
	ssize_t ret;
	
	memset(cmd + 4, 0x00, 4);
	cmd[0] =
	    CANOPEN_CSS_INITIATE_DOMAIN_UPLOAD;
	cmd[1] = index & 0xff;
	cmd[2] = index >> 8;
	cmd[3] = subindex & 0xff;

	pthread_mutex_lock(&dev->lock);

	ret = dev->send_request(dev, cmd, 8, &canopen_timeout);
	
	if (ret == 8)
		ret = dev->recv_response(dev, cmd, 8, &canopen_timeout);

	pthread_mutex_unlock(&dev->lock);

	if (ret != 8)
		return -1;
	
	if ((cmd[0] & CANOPEN_CSS_MASK) == CANOPEN_CSS_ABORT)
		return -1;

	if ((cmd[0] & CANOPEN_E) == 0)
		return 0;

	ret = 4;
	if (cmd[0] & CANOPEN_S)
		ret = 4 - ((cmd[0] & CANOPEN_N_MASK) >> CANOPEN_N_SHIFT);
	if (ret > bytes)
		ret = bytes;
	memcpy(data, cmd + 4, ret);
	
	return ret;
}

void canopen_emergency_handler(struct canopen_dev *dev, void *msg, size_t len)
{
	uint16_t code_le, code;
	uint8_t reg;
	uint8_t vendor[5];
	
	if (len != 8) {
		log_send(LOG_T_ERROR, 
		    "CANopen emergency received with wrong length: %zd",
		    len);
	} else {
		int log_type = LOG_T_ERROR;
		char *category = "unknown";
		char *cause = "unknown";
	
		memcpy(&code_le, msg, 2);
		code = le16toh(code_le);
		memcpy(&reg, msg + 2, 1);
		memcpy(vendor, msg + 3, 5);
		
		if (reg & 0x01)
			cause = "Generic Error";
		else if (reg & 0x02)
			cause = "Current";
		else if (reg & 0x04)
			cause = "Voltage";
		else if (reg & 0x08)
			cause = "Temperature";
		else if (reg & 0x10)
			cause = "Communication Error";
		else if (reg & 0x20)
			cause = "Device Profile Specific";
		else if (reg & 0x40)
			cause = "Reserved";
		else if (reg & 0x80)
			cause = "Manufacturer Specific";
		
		if (!(code & 0xff00)) {
			category = "Error reset";
			log_type = LOG_T_WARNING;
		} else if ((code & 0xff00) == 0x1000) {
			category = "Generic Error";
		} else if ((code & 0xf000) == 0x2000) {
			category = "Current";
		} else if ((code & 0xf000) == 0x3000) {
			category = "Voltage";
		} else if ((code & 0xf000) == 0x4000) {
			category = "Temperature";
		} else if ((code & 0xff00) == 0x5000) {
			category = "Device Hardware";
		} else if ((code & 0xff00) == 0x6000) {
			category = "Device Software";
		} else if ((code & 0xff00) == 0x7000) {
			category = "Additional Modules";
		} else if ((code & 0xff00) == 0x8000) {
			category = "Monitoring";
		} else if ((code & 0xff00) == 0x9000) {
			category = "External Error";
		} else if ((code & 0xff00) == 0xf000) {
			category = "Additional Functions";
		} else if ((code & 0xff00) == 0xff00) {
			category = "Device Specific";
		}
		
		log_send(log_type,
		    "%s: CANopen emergency message: "
		    "errorcode=0x%04x (%s), register=0x%02x (%s), "
		    "vendor=0x%02x%02x%02x%02x%02x",
		    dev->logname, code, category, reg, cause,
		    vendor[0], vendor[1], vendor[2], vendor[3], vendor[4]);
		
		if (dev->emergency_handler)
			dev->emergency_handler(dev, code, reg, vendor);
	}
}

void canopen_emergency_handler_attach(struct canopen_dev *dev,
    void (*func)(struct canopen_dev *dev, uint16_t code, uint8_t register, uint8_t vendor[5]))
{
	dev->emergency_handler = func;
}
    
int canopen_init(struct canopen_dev *dev)
{
	return pthread_mutex_init(&dev->lock, NULL);
}

