/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2009
	Copyright Stichting C.A. Muller Radioastronomiestation, 2009

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "ec.h"
#include "esc.h"
#include "esc_id.h"

void describe_esc(struct ec_dgram_addr *ec_addr)
{
	int r, r2;
	uint8_t val8;
	uint16_t val16;
	uint8_t eeprom[128*2];
	uint32_t vendorid, productcode;
	char *vendorid_str, *productcode_str;
	int i;
	
	ec_addr->addr.position.off = ESC_ADDR_MAP_TYPE;
	r = ec_datagram_read(ec_addr, &val8, 1);
	if (r == 1) {
		char *type;
		switch (val8) {
			case ESC_TYPE_ESC:
				type = "ESC";
				break;
			case ESC_TYPE_IPCORE:
				type = "IP core";
				break;
			case ESC_TYPE_ET1100:
				type = "ET1100";
				break;
			case ESC_TYPE_ET1200:
				type = "ET1200";
				break;
			default:
				type = "unknown";
		}
		printf("\tType: 0x%02x = %s\n", val8, type);
	}

	ec_addr->addr.position.off = ESC_ADDR_MAP_REVISION;
	r = ec_datagram_read(ec_addr, &val8, 1);

	ec_addr->addr.position.off = ESC_ADDR_MAP_BUILD;
	r2 = ec_datagram_read(ec_addr, &val16, 2);
	
	if (r == 1 && r2 == 2) {
		printf("\tRevision & Build: 0x%02x 0x%04x\n", val8, le16toh(val16));
	}

	ec_addr->addr.position.off = ESC_ADDR_MAP_FMMUS_SUPPORTED;
	r = ec_datagram_read(ec_addr, &val8, 1);
	
	if (r == 1) {
		printf("\tFMMUs supported: %d\n", val8);
	}
	
	ec_addr->addr.position.off = ESC_ADDR_MAP_SYNCMANAGERS_SUPPORTED;
	r = ec_datagram_read(ec_addr, &val8, 1);
	
	if (r == 1) {
		printf("\tSyncmanagers supported: %d\n", val8);
	}

	ec_addr->addr.position.off = ESC_ADDR_MAP_RAM_SIZE;
	r = ec_datagram_read(ec_addr, &val8, 1);
	
	if (r == 1) {
		printf("\tRAM size: %d Kb\n", val8);
	}
	
	ec_addr->addr.position.off = ESC_ADDR_MAP_FEATURES;
	r = ec_datagram_read(ec_addr, &val16, 2);
	
	if (r == 2) {
		val16 = le16toh(val16);
		printf("\tFeatures: 0x%02x\n", val16);
		if (val16 & ESC_FEATURE_FMMU_BYTE_ORIENTED)
			printf("\t\tFMMU is byte oriented\n");
		if (val16 & ESC_FEATURE_DISTRIBUTED_CLOCKS)
			printf("\t\tDistributed clocks\n");
		if (val16 & ESC_FEATURE_64BIT_DISTRIBUTED_CLOCKS)
			printf("\t\t64bit wide distributed clocks\n");
		if (val16 & ESC_FEATURE_LOW_JITTER_EBUS)
			printf("\t\tLow jitter EBUS\n");
		if (val16 & ESC_FEATURE_ENHANCED_LINK_DETECTION_EBUS)
			printf("\t\tEnhanced link detection EBUS\n");
		if (val16 & ESC_FEATURE_ENHANCED_LINK_DETECTION_MII)
			printf("\t\tEnhanced link detection MII\n");
		if (val16 & ESC_FEATURE_SEPERATE_FCS_HANDLING)
			printf("\t\tSeperate FCS handling\n");
		if (val16 & ESC_FEATURE_ENHANCED_DC_SYNC_ACTIVATION)
			printf("\t\tEnhanced DC sync activation\n");
	}

	r = esc_esi_eeprom_read(ec_addr, &eeprom, 0, sizeof(eeprom));
	if (r == sizeof(eeprom)) {
		printf("\tESI EEPROM:\n");
		for (i = 0; i < sizeof(eeprom)/sizeof(*eeprom); i += 16) {
			char *c;
			int j;
			
			printf("\t\t%02x: "
			    "%02x%02x %02x%02x %02x%02x %02x%02x "
			    "%02x%02x %02x%02x %02x%02x %02x%02x ",
			    i * 2,
			    eeprom[i+0], eeprom[i+1], eeprom[i+2], eeprom[i+3],
			    eeprom[i+4], eeprom[i+5], eeprom[i+6], eeprom[i+7],
			    eeprom[i+8], eeprom[i+9], eeprom[i+10], eeprom[i+11],
			    eeprom[i+12], eeprom[i+13], eeprom[i+14], eeprom[i+15]);
			c = (char*)&eeprom[i];
			for (j = 0; j < 16; j++) {
				if (isgraph(c[j])) {
					printf("%c", c[j]);
				} else {
					printf(".");
				}
			}
			printf("\n");
		}
	} else {
		printf("\tESI EEPROM read returned: %d expected: %zd.\n",
		    r, sizeof(eeprom));
	}
	
	vendorid = esc_esi_vendorid_get(ec_addr);
	productcode = esc_esi_productcode_get(ec_addr);
	
	vendorid_str = "unknown vendorid";
	productcode_str = "unkown productcode";
	for (i = 0; i < esc_vendor_list_len; i++)
		if (esc_vendor_list[i].vendorid == vendorid)
			vendorid_str = esc_vendor_list[i].name;
	for (i = 0; i < esc_device_list_len; i++)
		if (esc_device_list[i].vendorid == vendorid &&
		    esc_device_list[i].productcode == productcode)
			productcode_str = esc_device_list[i].name;
	
	printf("\tESI VendorId: 0x%04x\n", vendorid);
	printf("\t\tVendor: %s\n", vendorid_str);
	printf("\tESI ProductCode: 0x%04x\n", productcode);
	printf("\t\tProduct: %s\n", productcode_str);
	printf("\tESI RevisionNo: 0x%04x\n", esc_esi_revisionno_get(ec_addr));
	printf("\tESI SerialNo: 0x%04x\n", esc_esi_serialno_get(ec_addr));
	
	printf("\n");
}

int main(int argc, char **argv)
{
	int slaves;
	int i;
	
	if (argc < 2) {
		printf("Usage: %s [ethernet device]\n", argv[0]);
		return -1;
	}
	
	if (ec_init(argv[1], 0)) {
		printf("EtherCat init failed\n");
		return -1;
	}
	
	slaves = ec_slave_count();
	
	printf("Number of slaves found: %d\n", slaves);
	
	for (i = 0; i < slaves; i++) {
		struct ec_dgram_addr ec_addr;
		
		printf("EtherCat slave controller %d:\n", i);
		
		ec_addr_set_auto_inc_nr(&ec_addr, i);
		
		describe_esc(&ec_addr);
	}
	
	return 0;
}
