/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2009
	Copyright Stichting C.A. Muller Radioastronomiestation, 2009   

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <string.h>
#include <stdio.h>
#include <time.h>
#include <pthread.h>
#include <limits.h>

#include "esc.h"
#include "ec.h"
#include "canopen.h"
#include "log.h"

struct esc_mailbox {
	struct ec_dgram_addr status_rd;
	struct ec_dgram_addr data_rd;
	struct ec_dgram_addr status_wr;
	struct ec_dgram_addr data_wr;
	int ctr;
	int len_rd;
	int len_wr;
	unsigned char *buffer_rd;
	unsigned char *buffer_wr;
	
	void (*error_handler)(void *msg, size_t len);
};


struct esc_mailbox_hdr {
	uint16_t length;
	uint16_t address;
	uint8_t chan_prio;
	uint8_t type_ctr;
} __attribute((__packed__));

#define ESC_MAILBOX_TYPE_ERROR	0x0
#define ESC_MAILBOX_TYPE_ADS	0x1
#define ESC_MAILBOX_TYPE_EOE	0x2
#define ESC_MAILBOX_TYPE_COE	0x3
#define ESC_MAILBOX_TYPE_FOE	0x4
#define ESC_MAILBOX_TYPE_SOE	0x5
#define ESC_MAILBOX_TYPE_VOE	0xf

#define ESC_COE_TIMEOUT	2

ssize_t esc_esi_eeprom_read(struct ec_dgram_addr *ec_addr, 
    void *buffer, size_t offset, size_t size)
{
	int r;
	uint8_t val8;
	uint16_t val16, status;
	uint32_t val32;
	int wordsize;
	size_t readsize = 0;
	
	/* Make sure ecat side has eeprom control */
	do {
		ec_addr->addr.position.off = ESC_ADDR_MAP_ESI_EEPROM_CONFIG;
		r = ec_datagram_read(ec_addr, &val8, 1);
		if (val8 & ESC_ESI_EEPROM_CONFIG_PDI && r == 1) {
			uint8_t resetval8 = ESC_ESI_EEPROM_CONFIG_FORCE_ECAT;
			r = ec_datagram_write(ec_addr, &resetval8, 1);
		}
	} while (val8 & ESC_ESI_EEPROM_CONFIG_PDI && r == 1);
	if (r != 1)
		return -1;

	/* get status */
	ec_addr->addr.position.off = ESC_ADDR_MAP_ESI_EEPROM_CONTROL_STATUS;
	r = ec_datagram_read(ec_addr, &val16, 2);
	if (r != 2)
		return -1;
	
	status = le16toh(val16);
	if (status & ESC_ESI_EEPROM_CONTROL_STATUS_READ_8B) {
		wordsize = 8;
	} else {
		wordsize = 4;
	}

	while (size) {
		uint8_t data[8];
		size_t copysize;
		size_t copyoffset;
		
		/* set address */
		val32 = htole32(offset / (wordsize / 2));
		ec_addr->addr.position.off = ESC_ADDR_MAP_ESI_EEPROM_ADDRESS;
		r = ec_datagram_write(ec_addr, &val32, 4);
		
		/* Wait for other stuff to finish */
		do {
			r = ec_datagram_read(ec_addr, &val16, 2);
		} while (le16toh(val16) & ESC_ESI_EEPROM_CONTROL_STATUS_BUSY && r == 2);

		/* Issue read command */
		ec_addr->addr.position.off = ESC_ADDR_MAP_ESI_EEPROM_CONTROL_STATUS;
		val16 = htole16(ESC_ESI_EEPROM_CONTROL_STATUS_CMD_READ);
		r = ec_datagram_write(ec_addr, &val16, 2);
		
		/* Wait for read to finish */
		do {
			r = ec_datagram_read(ec_addr, &val16, 2);
		} while (le16toh(val16) & ESC_ESI_EEPROM_CONTROL_STATUS_BUSY && r == 2);
		
		/* make sure nothing is fucked up.... */
		status = le16toh(val16);
		if (status & ESC_ESI_EEPROM_CONTROL_STATUS_CHKSUM_ERR ||
		    status & ESC_ESI_EEPROM_CONTROL_STATUS_NOT_LOADED ||
		    status & ESC_ESI_EEPROM_CONTROL_STATUS_ACK_ERR ||
		    status & ESC_ESI_EEPROM_CONTROL_STATUS_WRITE_ERR) {
			return -1;
		}

		/* read data and copy to client buffer */
		ec_addr->addr.position.off = ESC_ADDR_MAP_ESI_EEPROM_DATA;
		r = ec_datagram_read(ec_addr, data, wordsize);
		if (r != wordsize)
			break;
		
		copysize = size;
		if (copysize > wordsize)
			copysize = wordsize;
		copyoffset = offset & (wordsize -1);
		if (copysize + copyoffset > wordsize)
			copysize = wordsize - offset;
		memcpy(buffer, &data[copyoffset], copysize);
		size -= copysize;
		offset += copysize;
		buffer += copysize;
		readsize += copysize;
	}
	
	return readsize;
}


static uint32_t esc_esi_read32(struct ec_dgram_addr *ec_addr, size_t offset)
{
	uint32_t val32;
	
	esc_esi_eeprom_read(ec_addr, &val32, offset, sizeof(uint32_t));
	
	return le32toh(val32);
}

uint32_t esc_esi_vendorid_get(struct ec_dgram_addr *ec_addr)
{
	return esc_esi_read32(ec_addr, ESC_ESI_ADDR_VENDORID);
}

uint32_t esc_esi_productcode_get(struct ec_dgram_addr *ec_addr)
{
	return esc_esi_read32(ec_addr, ESC_ESI_ADDR_PRODUCTCODE);
}

uint32_t esc_esi_revisionno_get(struct ec_dgram_addr *ec_addr)
{
	return esc_esi_read32(ec_addr, ESC_ESI_ADDR_REVISIONNO);
}

uint32_t esc_esi_serialno_get(struct ec_dgram_addr *ec_addr)
{
	return esc_esi_read32(ec_addr, ESC_ESI_ADDR_SERIALNO);
}

static bool esc_timeout_expired(struct timespec *start, struct timespec *timeout)
{
	time_t sec;
	long nanosec;
	struct timespec t_now;
	
	sec = start->tv_sec + timeout->tv_sec;
	nanosec = start->tv_nsec + timeout->tv_nsec;
	
	if (nanosec > 1000000000) {
		sec++;
		nanosec -= 1000000000;
	}
	clock_gettime(CLOCK_MONOTONIC, &t_now);
	
	/* seconds expired? */
	if (t_now.tv_sec < sec)
		return false;
	if (t_now.tv_sec > sec)
		return true;

	/* within one second, check nanoseconds*/
	if (t_now.tv_nsec > nanosec)
		return true;
	
	return false;
}

int esc_al_state_get(struct ec_dgram_addr *addr)
{
	uint16_t state;
	struct ec_dgram_addr state_add;
	int ret;
	
	memcpy(&state_add, addr, sizeof(state_add));
	state_add.addr.position.off = ESC_ADDR_MAP_AL_STATUS;
	ret = ec_datagram_read(&state_add, &state, 2);

	if (ret == 2) {
		return le16toh(state) & 0x000f;
	} else
		return -1;
}

int esc_al_state_set(struct ec_dgram_addr *addr, enum esc_al_state newstate,
    struct timespec *timeout)
{
	uint16_t state;
	struct ec_dgram_addr state_add;
	int ret;
	struct timespec t_start;
	
	memcpy(&state_add, addr, sizeof(state_add));
	state_add.addr.position.off = ESC_ADDR_MAP_AL_CONTROL;
	state = htole16(newstate);
	do {
		ret = ec_datagram_write(&state_add, &state, sizeof(uint16_t));
	} while (ret == 0);
	
	log_send(LOG_T_DEBUG, "set state: %d datagram write: %d", newstate, ret);
	if (ret == 2) {
		clock_gettime(CLOCK_MONOTONIC, &t_start);
		while (!esc_timeout_expired(&t_start, timeout)) {
			int status;
			
			state = esc_al_state_get(addr);
			if (state == newstate)
				return newstate;
			else
				log_send(LOG_T_DEBUG, "state: %d != %d",
				    state, newstate);
			status = esc_al_status_code_get(addr);
			if (status != 0) {
				log_send(LOG_T_DEBUG, 
				    "state now: %d status code: 0x%x",
				    state, status);
				esc_al_error_ack(addr);
				
				return -1;
			}
		}
		log_send(LOG_T_ERROR, 
		    "Timeout while trying to set state %d, stuck in state %d",
		    newstate, state);
		/* timeout... */
		return state;
	} else
		return -1;
}

int esc_al_status_code_get(struct ec_dgram_addr *addr)
{
	uint16_t state;
	struct ec_dgram_addr state_add;
	int ret;
	
	memcpy(&state_add, addr, sizeof(state_add));
	state_add.addr.position.off = ESC_ADDR_MAP_AL_STATUS;
	ret = ec_datagram_read(&state_add, &state, 2);
	if (ret != 2)
		goto err_get_state;
	
	if (le16toh(state) & ESC_ADDR_MAP_AL_STATUS_ERR_IND) {
		state_add.addr.position.off = ESC_ADDR_MAP_AL_STATUS_CODE;
		ret = ec_datagram_read(&state_add, &state, 2);
		state = le16toh(state);
	} else {
		state = 0;
	}
	
	return state;

err_get_state:
	return -1;
}

int esc_al_error_ack(struct ec_dgram_addr *addr)
{
	uint16_t ctrl;
	struct ec_dgram_addr state_add;
	int ret;
	
	memcpy(&state_add, addr, sizeof(state_add));

	ret = esc_al_state_get(&state_add);
	if (ret < 0)
		return -1;
	ctrl = ret;
	
	state_add.addr.position.off = ESC_ADDR_MAP_AL_CONTROL;
	ctrl |= htole16(ESC_ADDR_MAP_AL_CONTROL_ERROR_ACK);
	ret = ec_datagram_write(&state_add, &ctrl, 2);
	
	if (ret == 2)
		return 0;
	else
		return -1;
}

int esc_init(struct ec_dgram_addr *addr)
{
	uint8_t nr_syncmgr;
	int ret, i;

	addr->addr.position.off = ESC_ADDR_MAP_SYNCMANAGERS_SUPPORTED;
	ret = ec_datagram_read(addr, &nr_syncmgr, 1);
	if (ret != 1) {
		log_send(LOG_T_ERROR, "Could not read ESC information");
		return -1;
	}

	log_send(LOG_T_DEBUG, "ESC supports %d SyncManagers", nr_syncmgr);
	/* De-activate all sync managers. */
	for (i = 0; i < nr_syncmgr; i++) {
		uint16_t off = ESC_ADDR_MAP_SYNCMANAGER_NR(i);
		uint8_t val8;
		uint16_t val16;
		
		addr->addr.position.off = off | ESC_ADDR_MAP_SYNCMANAGER_ACTIVATE;
		val8 = 0;
		
		ret = ec_datagram_write(addr, &val8, 1);
		if (ret != 1) {
			log_send(LOG_T_ERROR, 
			    "Could not deactivate sync manager %d", i);
			return -1;
		}

		val16 = 0;
		addr->addr.position.off = off | ESC_ADDR_MAP_SYNCMANAGER_LENGTH;
			ec_datagram_write(addr, &val16, 2);
	}

	return 0;
}

int esc_syncmanager_claim(struct ec_dgram_addr *addr)
{
	int ret, i;
	int nr = -1;
	uint8_t nr_syncmgr;

	addr->addr.position.off = ESC_ADDR_MAP_SYNCMANAGERS_SUPPORTED;
	ret = ec_datagram_read(addr, &nr_syncmgr, 1);
	if (ret != 1) {
		log_send(LOG_T_ERROR, "Could not read ESC information");
		return -1;
	}

	for (i = 0; i < nr_syncmgr; i++) {
		uint16_t off = ESC_ADDR_MAP_SYNCMANAGER_NR(i);
		uint8_t val8;
		
		addr->addr.position.off = off | ESC_ADDR_MAP_SYNCMANAGER_ACTIVATE;
		
		ret = ec_datagram_read(addr, &val8, 1);
		if (ret != 1) {
			log_send(LOG_T_ERROR, 
			    "Could not read syncmanger%d activate register", i);
			return -1;
		}
		if (!(val8 & ESC_SYNCMANAGER_ACTIVATE_ENABLE)) {
			uint16_t val16;

			addr->addr.position.off = off |
			    ESC_ADDR_MAP_SYNCMANAGER_LENGTH;
			ec_datagram_read(addr, &val16, 2);
			if (val16 == 0) {
				nr = i;
				break;
			}
		}
	}

	return nr;
}

static void esc_error_handler_default(void *msg, size_t len)
{
	unsigned char *msgb = msg;
	int i;
	char msgstr[100 + len];
	int pos = 0;
	
	pos = sprintf(msgstr, "Mailbox error message:");
	for (i = 0; i < len; i++) {
		pos += sprintf(msgstr + pos, " 0x%02x", msgb[i]);
	}
	log_send(LOG_T_ERROR, msgstr);
}

struct esc_mailbox *esc_mailbox_create(struct ec_dgram_addr *node,
    struct esc_syncmanager_info *rd, struct esc_syncmanager_info *wr)
{
	int sync_manager_write, sync_manager_read;
	size_t start_write, start_read;
	size_t len_write, len_read;
	struct esc_mailbox mb, *mailbox = NULL;
	uint8_t val8;
	uint16_t leval16;
	
	/* init write sync manager */
	if (!wr) {
		sync_manager_write = esc_syncmanager_claim(node);
		if (sync_manager_write < 0)
			return NULL;
		start_write = 0x1000;
		len_write = 0x0100;
	} else {
		sync_manager_write = wr->sm;
		start_write = wr->start;
		len_write = wr->len;
	}
	
	log_send(LOG_T_DEBUG, "Claiming syncmanager %d for mailbox write",
	    sync_manager_write);
	
	/* mailbox write (master->slave) */	

	node->addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_write) |
	    ESC_ADDR_MAP_SYNCMANAGER_ACTIVATE;
	val8 = 0x00;
	ec_datagram_write(node, &val8, 1);
	
	node->addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_write) |
	    ESC_ADDR_MAP_SYNCMANAGER_PHYSICAL_START;
	leval16 = htole16(start_write);
	ec_datagram_write(node, &leval16, 2);
	log_send(LOG_T_DEBUG, "write address: %04zx", start_write);
	node->addr.position.off =
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_write) | 
	    ESC_ADDR_MAP_SYNCMANAGER_LENGTH;
	leval16 = htole16(len_write);
	ec_datagram_write(node, &leval16, 2);
	
	node->addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_write) | 
	    ESC_ADDR_MAP_SYNCMANAGER_CONTROL;
	val8 = 
	    ESC_SYNCMANAGER_CONTROL_MAILBOX | 
	    ESC_SYNCMANAGER_CONTROL_WRITE |
	    ESC_SYNCMANAGER_CONTROL_INT_PDI;
	log_send(LOG_T_DEBUG, "Write Control byte: 0x%02x", val8);
	ec_datagram_write(node, &val8, 1);

	node->addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_write) | 
	    ESC_ADDR_MAP_SYNCMANAGER_STATUS;
	ec_datagram_read(node, &val8, 1);
	log_send(LOG_T_DEBUG, "Write status byte: 0x%02x", val8);
	
	node->addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_write) | 
	    ESC_ADDR_MAP_SYNCMANAGER_ACTIVATE;
	val8 = ESC_SYNCMANAGER_ACTIVATE_ENABLE;
	ec_datagram_write(node, &val8, 1);
 

	/* init read sync manager */ 
	if (!rd) {
		sync_manager_read = esc_syncmanager_claim(node);
		if (sync_manager_read < 0)
			return NULL;
		start_read = 0x1100;
		len_read = 0x0100;
	} else {
		sync_manager_read = rd->sm;
		start_read = rd->start;
		len_read = rd->len;
	}
	
	log_send(LOG_T_ERROR, "Claiming syncmanager %d for mailbox read",
	    sync_manager_read);
	
	/* mailbox read (slave->master)  */
	
	node->addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_read) | 
	    ESC_ADDR_MAP_SYNCMANAGER_ACTIVATE;
	val8 = 0x00;
	ec_datagram_write(node, &val8, 1);
	
	node->addr.position.off =
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_read) | 
	    ESC_ADDR_MAP_SYNCMANAGER_PHYSICAL_START;
	leval16 = htole16(start_read);
	ec_datagram_write(node, &leval16, 2);
	log_send(LOG_T_DEBUG, "read address: %04zx", start_read);
	
	node->addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_read) | 
	    ESC_ADDR_MAP_SYNCMANAGER_LENGTH;
	leval16 = htole16(len_read);
	ec_datagram_write(node, &leval16, 2);
	
	node->addr.position.off =
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_read) | 
	    ESC_ADDR_MAP_SYNCMANAGER_CONTROL;
	val8 = 
	    ESC_SYNCMANAGER_CONTROL_MAILBOX | 
	    ESC_SYNCMANAGER_CONTROL_INT_PDI;
	log_send(LOG_T_DEBUG, "Read Control byte: 0x%02x", val8);
	ec_datagram_write(node, &val8, 1);
	

	node->addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_read) | 
	    ESC_ADDR_MAP_SYNCMANAGER_STATUS;
	ec_datagram_read(node, &val8, 1);
	log_send(LOG_T_DEBUG, "Read status byte: 0x%02x", val8);

	node->addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_read) | 
	    ESC_ADDR_MAP_SYNCMANAGER_ACTIVATE;
	val8 = ESC_SYNCMANAGER_ACTIVATE_ENABLE;
	ec_datagram_write(node, &val8, 1);

	memcpy(&mb.status_rd, node, sizeof(struct ec_dgram_addr));
	mb.status_rd.addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_read) | 
	    ESC_ADDR_MAP_SYNCMANAGER_STATUS;
	memcpy(&mb.status_wr, node, sizeof(struct ec_dgram_addr));
	mb.status_wr.addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_write) | 
	    ESC_ADDR_MAP_SYNCMANAGER_STATUS;
	mb.len_rd = len_read;
	memcpy(&mb.data_rd, node, sizeof(struct ec_dgram_addr));
	mb.data_rd.addr.position.off = start_read;
	mb.len_wr = len_write;
	memcpy(&mb.data_wr, node, sizeof(struct ec_dgram_addr));
	mb.data_wr.addr.position.off = start_write;
	
	mailbox = malloc(sizeof(struct esc_mailbox));
	if (mailbox) {
		memcpy(mailbox, &mb, sizeof(struct esc_mailbox));
		if (mailbox->len_rd) {
			mailbox->buffer_rd = malloc(mailbox->len_rd);
			if (!mailbox->buffer_rd)
				goto err_buffer_rd;
		} else {
			mailbox->buffer_rd = NULL;
		}
		if (mailbox->len_wr) {
			mailbox->buffer_wr = malloc(mailbox->len_wr);
			if (!mailbox->buffer_wr)
				goto err_buffer_wr;
		}
		mailbox->error_handler = esc_error_handler_default;
	}
	
	return mailbox;

err_buffer_wr:
	if (mailbox->buffer_rd)
		free(mailbox->buffer_rd);
err_buffer_rd:
	free(mailbox);
	return NULL;
}

void esc_mailbox_reinit(struct ec_dgram_addr *node,
    struct esc_syncmanager_info *rd, struct esc_syncmanager_info *wr)
{
	int sync_manager_write, sync_manager_read;
	size_t start_write, start_read;
	size_t len_write, len_read;
	struct esc_mailbox mb;
	uint8_t val8;
	uint16_t leval16;
	
	/* init write sync manager */
	if (!wr) {
		sync_manager_write = esc_syncmanager_claim(node);
		if (sync_manager_write < 0)
			return;
		start_write = 0x1000;
		len_write = 0x0100;
	} else {
		sync_manager_write = wr->sm;
		start_write = wr->start;
		len_write = wr->len;
	}
	
	log_send(LOG_T_DEBUG, "Claiming syncmanager %d for mailbox write",
	    sync_manager_write);
	
	/* mailbox write (master->slave) */	

	node->addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_write) |
	    ESC_ADDR_MAP_SYNCMANAGER_ACTIVATE;
	val8 = 0x00;
	ec_datagram_write(node, &val8, 1);
	
	node->addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_write) |
	    ESC_ADDR_MAP_SYNCMANAGER_PHYSICAL_START;
	leval16 = htole16(start_write);
	ec_datagram_write(node, &leval16, 2);
	log_send(LOG_T_DEBUG, "write address: %04zx", start_write);
	node->addr.position.off =
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_write) | 
	    ESC_ADDR_MAP_SYNCMANAGER_LENGTH;
	leval16 = htole16(len_write);
	ec_datagram_write(node, &leval16, 2);
	
	node->addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_write) | 
	    ESC_ADDR_MAP_SYNCMANAGER_CONTROL;
	val8 = 
	    ESC_SYNCMANAGER_CONTROL_MAILBOX | 
	    ESC_SYNCMANAGER_CONTROL_WRITE |
	    ESC_SYNCMANAGER_CONTROL_INT_PDI;
	log_send(LOG_T_DEBUG, "Write Control byte: 0x%02x", val8);
	ec_datagram_write(node, &val8, 1);

	node->addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_write) | 
	    ESC_ADDR_MAP_SYNCMANAGER_STATUS;
	ec_datagram_read(node, &val8, 1);
	log_send(LOG_T_DEBUG, "Write status byte: 0x%02x", val8);
	
	node->addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_write) | 
	    ESC_ADDR_MAP_SYNCMANAGER_ACTIVATE;
	val8 = ESC_SYNCMANAGER_ACTIVATE_ENABLE;
	ec_datagram_write(node, &val8, 1);
 

	/* init read sync manager */ 
	if (!rd) {
		sync_manager_read = esc_syncmanager_claim(node);
		if (sync_manager_read < 0)
			return;
		start_read = 0x1100;
		len_read = 0x0100;
	} else {
		sync_manager_read = rd->sm;
		start_read = rd->start;
		len_read = rd->len;
	}
	
	log_send(LOG_T_DEBUG, "Claiming syncmanager %d for mailbox read",
	    sync_manager_read);
	
	/* mailbox read (slave->master)  */
	
	node->addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_read) | 
	    ESC_ADDR_MAP_SYNCMANAGER_ACTIVATE;
	val8 = 0x00;
	ec_datagram_write(node, &val8, 1);
	
	node->addr.position.off =
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_read) | 
	    ESC_ADDR_MAP_SYNCMANAGER_PHYSICAL_START;
	leval16 = htole16(start_read);
	ec_datagram_write(node, &leval16, 2);
	log_send(LOG_T_DEBUG, "read address: %04zx", start_read);
	
	node->addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_read) | 
	    ESC_ADDR_MAP_SYNCMANAGER_LENGTH;
	leval16 = htole16(len_read);
	ec_datagram_write(node, &leval16, 2);
	
	node->addr.position.off =
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_read) | 
	    ESC_ADDR_MAP_SYNCMANAGER_CONTROL;
	val8 = 
	    ESC_SYNCMANAGER_CONTROL_MAILBOX | 
	    ESC_SYNCMANAGER_CONTROL_INT_PDI;
	log_send(LOG_T_DEBUG, "Read Control byte: 0x%02x", val8);
	ec_datagram_write(node, &val8, 1);
	

	node->addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_read) | 
	    ESC_ADDR_MAP_SYNCMANAGER_STATUS;
	ec_datagram_read(node, &val8, 1);
	log_send(LOG_T_DEBUG, "Read status byte: 0x%02x", val8);

	node->addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_read) | 
	    ESC_ADDR_MAP_SYNCMANAGER_ACTIVATE;
	val8 = ESC_SYNCMANAGER_ACTIVATE_ENABLE;
	ec_datagram_write(node, &val8, 1);

	memcpy(&mb.status_rd, node, sizeof(struct ec_dgram_addr));
	mb.status_rd.addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_read) | 
	    ESC_ADDR_MAP_SYNCMANAGER_STATUS;
	memcpy(&mb.status_wr, node, sizeof(struct ec_dgram_addr));
	mb.status_wr.addr.position.off = 
	    ESC_ADDR_MAP_SYNCMANAGER_NR(sync_manager_write) | 
	    ESC_ADDR_MAP_SYNCMANAGER_STATUS;
	mb.len_rd = len_read;
	memcpy(&mb.data_rd, node, sizeof(struct ec_dgram_addr));
	mb.data_rd.addr.position.off = start_read;
	mb.len_wr = len_write;
	memcpy(&mb.data_wr, node, sizeof(struct ec_dgram_addr));
	mb.data_wr.addr.position.off = start_write;
	
}

void esc_mailbox_destroy(struct esc_mailbox *mailbox)
{
	/* todo: deactivate sync managers */
	
	free(mailbox->buffer_rd);
	free(mailbox->buffer_wr);
	free(mailbox);
}

ssize_t esc_mailbox_write(struct esc_mailbox *mailbox, int type,
    void *data, size_t len, struct timespec *timeout)
{
	struct esc_mailbox_hdr *hdr;
	uint8_t val8;
	int ret;
	struct timespec t_start;
	
	if (mailbox->len_wr == 0) {
		log_send(LOG_T_ERROR, "len_wr == 0");
		return -1;
	}
	if (mailbox->len_wr < sizeof(struct esc_mailbox_hdr) + len ) {
		log_send(LOG_T_ERROR, "mailbox->len_wr to small: %d < %zu",
		    mailbox->len_wr, sizeof(struct esc_mailbox_hdr) + len);
		return -1;
	}
		
	hdr = (struct esc_mailbox_hdr *)mailbox->buffer_wr;
	hdr->length = htole16(len);
	hdr->address = htole16(0x0);
	hdr->chan_prio = 0;
	hdr->type_ctr = type | (mailbox->ctr << 4);
	memcpy(mailbox->buffer_wr + sizeof(struct esc_mailbox_hdr), data, len);

	mailbox->ctr++;
	
	/* wait till mailbox is empty */
	clock_gettime(CLOCK_MONOTONIC, &t_start);
	do {
		ret = ec_datagram_read(&mailbox->status_wr, &val8,
		    sizeof(val8));
		if (ret != sizeof(val8)) {
			log_send(LOG_T_ERROR, "could not read mailbox status");
			return -1;
		}
	} while ((val8 & 0x08) && !esc_timeout_expired(&t_start, timeout));

	if (!(val8 & 0x08)) {
		/* write mailbox contents */
		ret = ec_datagram_write(&mailbox->data_wr, mailbox->buffer_wr,
		    mailbox->len_wr);
		if (ret != mailbox->len_wr) {
			return -1;
		}
		ret = len;
	} else {
		ret = 0;
	}
	
	return ret;
}

ssize_t esc_mailbox_read(struct esc_mailbox *mailbox, int *type,
    void *data, size_t len, struct timespec *timeout)
{
	struct esc_mailbox_hdr *hdr;
	int ret;
	size_t clen;
	uint8_t val8;
	struct timespec t_start;
		
	if (mailbox->len_rd == 0) {
		log_send(LOG_T_ERROR, "len_rd == 0");
		return -1;
	}
	if (mailbox->len_rd < sizeof(struct esc_mailbox_hdr) + len ) {
		log_send(LOG_T_ERROR, "len_rd to short %d < %zu",
		    mailbox->len_rd, sizeof(struct esc_mailbox_hdr) + len);
		return -1;
	}
	
	hdr = (struct esc_mailbox_hdr *)mailbox->buffer_rd;

	/* wait untill mailbox full */
	clock_gettime(CLOCK_MONOTONIC, &t_start);
	do {
		ret = ec_datagram_read(&mailbox->status_rd, &val8, sizeof(val8));
		if (ret != sizeof(val8)) {
			log_send(LOG_T_ERROR, "Could not read mailbox status");
			return -1;
		}
	} while (!(val8 & 0x08) && !esc_timeout_expired(&t_start, timeout));

	if (!(val8 & 0x08)) {
		return 0;
	}

	/* read mailbox contents */
	ret = ec_datagram_read(&mailbox->data_rd, mailbox->buffer_rd,
	    mailbox->len_rd);
	if (ret != mailbox->len_rd) {
		return -1;
	}

	clen = le16toh(hdr->length);
	if (clen > len)
		clen = len;
	
	memcpy(data, mailbox->buffer_rd + sizeof(struct esc_mailbox_hdr), clen);
	*type = hdr->type_ctr & 0x0f;
	
	if (*type == ESC_MAILBOX_TYPE_ERROR) {
		mailbox->error_handler(
		    mailbox->buffer_rd + sizeof(struct esc_mailbox_hdr),
		    hdr->length);
	}

	return clen;
}


struct canopen_dev_priv {
	struct esc_mailbox *mbox;
	int refcount;
};

static ssize_t esc_coe_send(struct canopen_dev *dev, void *req, size_t len,
    struct timespec *timeout)
{
	int ret;
	unsigned char coereq[10];
	int coe_type = 0x2;
	
	memcpy(coereq + 2, req, len);
	coereq[0] = 0x00;
	coereq[1] = coe_type << 4;
	
	ret = esc_mailbox_write(dev->priv->mbox, ESC_MAILBOX_TYPE_COE, coereq, 
	    len+2, timeout);

	if (ret > 2)
		return ret - 2;
	return ret;
}

static ssize_t esc_coe_recv(struct canopen_dev *dev, void *res, size_t len,
    struct timespec *timeout)
{
	int type;
	int ret;
	unsigned char coeres[10];
	int coe_type;
	
	ret = esc_mailbox_read(dev->priv->mbox, &type, coeres, len + 2, timeout);
	if (ret == 0)
		return 0;
	if (type != ESC_MAILBOX_TYPE_COE) {
		return -1;
	}
	
	coe_type = (coeres[1] >> 4) & 0x0f;

	/* check for emergencies */
	if (coe_type == 1) {
		canopen_emergency_handler(dev, coeres + 2, ret - 2);
		return 0;
	}
	
	if (ret > 2) {
		memcpy(res, coeres + 2, ret - 2);
		return ret - 2;
	}
	return ret;
}

static void *esc_coe_emergency_poll(void *arg)
{
	struct canopen_dev *dev = arg;
	unsigned char cmd[8];
	struct timespec timeout = { 0, 0 };
	
	log_send(LOG_T_DEBUG, "Emergency poll thread started for CoE device %p",
	    dev);
	while (1) {
		sleep(1);

		pthread_mutex_lock(&dev->lock);
	
		dev->recv_response(dev, cmd, 8, &timeout);
	
		pthread_mutex_unlock(&dev->lock);
	}
	return NULL;
}

struct canopen_dev *esc_coe_create(struct esc_mailbox *mailbox)
{
	struct canopen_dev *dev;
	pthread_t thread_id;
	pthread_attr_t attr;
	
	dev = malloc(sizeof(struct canopen_dev));
	if (!dev)
		goto err_malloc;
	dev->priv = malloc(sizeof(struct canopen_dev_priv));
	if (!dev->priv)
		goto err_priv;
	
	dev->priv->mbox = mailbox;
	dev->send_request = &esc_coe_send;
	dev->recv_response = &esc_coe_recv;
	dev->logname = "esc";
	dev->priv->refcount = 2;

	canopen_init(dev);

	pthread_attr_init(&attr);
	pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN * 2);
	pthread_create(&thread_id, &attr,
	    esc_coe_emergency_poll, dev);

	return dev;

err_priv:
	free(dev);
err_malloc:
	return NULL;
}

void esc_coe_destroy(struct canopen_dev *coe)
{
	coe->priv->refcount--;

	if (!coe->priv->refcount) {
		free(coe->priv);
		free(coe);
	}
}

int esc_dc_init(struct ec_dgram_addr *addr)
{
	uint8_t val8;
	uint64_t leval64;
	struct timespec t;
	
	/* Configure distributed clock and sync0 (4ms cycle time) */
	addr->addr.position.off = ESC_ADDR_MAP_DC_CYCLIC_UNIT_CONTROL;
	val8 = 0x00;
	ec_datagram_write(addr, &val8, 1);

	addr->addr.position.off = ESC_ADDR_MAP_DC_ACTIVATION_REGISTER;
	val8 = 
	    ESC_DC_ACTIVATION_REGISTER_SYNC_OUT_ACTIVE |
	    ESC_DC_ACTIVATION_REGISTER_SYNC0_ACTIVE;
	ec_datagram_write(addr, &val8, 1);


	addr->addr.position.off = ESC_ADDR_MAP_DC_START_TIME_CYCLIC_OPERATION;
	clock_gettime(0, &t);
	leval64 = htole64((ESC_DC_EPOCH + (uint64_t)t.tv_sec) * 1000*1000*1000);
	ec_datagram_write(addr, &leval64, 8);

	return 0;
}

int esc_dc_sync0_cycle_time(struct ec_dgram_addr *addr, unsigned long ns)
{
	int r;
	uint32_t leval32;
	
	addr->addr.position.off = ESC_ADDR_MAP_DC_SYNC0_CYCLE_TIME;
	leval32 = htole32(ns);
	r = ec_datagram_write(addr, &leval32, 4);

	if (r == 4)
		return 0;
	else
		return -1;
}

int esc_pdo_tx_set(struct ec_dgram_addr *addr,
    struct esc_syncmanager_info *sm, unsigned char **data,
    bool watchdog)
{
	uint8_t val8;
	uint16_t leval16;

	addr->addr.position.off = ESC_ADDR_MAP_SYNCMANAGER_NR(sm->sm) |
	   ESC_ADDR_MAP_SYNCMANAGER_ACTIVATE;
	val8 = 0x00;
	ec_datagram_write(addr, &val8, 1);

	addr->addr.position.off = ESC_ADDR_MAP_SYNCMANAGER_NR(sm->sm) |
	    ESC_ADDR_MAP_SYNCMANAGER_PHYSICAL_START;
	leval16= htole16(sm->start);
	ec_datagram_write(addr, &leval16, 2);
	
	/* Length */
	addr->addr.position.off = ESC_ADDR_MAP_SYNCMANAGER_NR(sm->sm) |
	    ESC_ADDR_MAP_SYNCMANAGER_LENGTH;
	leval16 = htole16(sm->len);
	ec_datagram_write(addr, &leval16, 2);
	
	addr->addr.position.off = ESC_ADDR_MAP_SYNCMANAGER_NR(sm->sm) |
	    ESC_ADDR_MAP_SYNCMANAGER_CONTROL;
	val8 = 
	    ESC_SYNCMANAGER_CONTROL_WRITE |
	    ESC_SYNCMANAGER_CONTROL_INT_PDI;
	if (watchdog)
		val8 |= ESC_SYNCMANAGER_CONTROL_WATCHDOG_ENABLE;
	log_send(LOG_T_DEBUG, "PDO tx Control byte: 0x%02x", val8);
	ec_datagram_write(addr, &val8, 1);
	

	if (sm->len) {
		addr->addr.position.off = ESC_ADDR_MAP_SYNCMANAGER_NR(sm->sm) |
		    ESC_ADDR_MAP_SYNCMANAGER_ACTIVATE;
		val8 = ESC_SYNCMANAGER_ACTIVATE_ENABLE;
		ec_datagram_write(addr, &val8, 1);

		if (data)
		{
			addr->addr.position.off = sm->start;
			ec_tx_pdo_add(addr, data, sm->len);
		}
	}

	return 0;
}

int esc_pdo_rx_set(struct ec_dgram_addr *addr,
    struct esc_syncmanager_info *sm, unsigned char **data)
{
	uint8_t val8;
	uint16_t leval16;
	
	addr->addr.position.off = ESC_ADDR_MAP_SYNCMANAGER_NR(sm->sm) |
	   ESC_ADDR_MAP_SYNCMANAGER_ACTIVATE;
	val8 = 0x00;
	ec_datagram_write(addr, &val8, 1);

	addr->addr.position.off = ESC_ADDR_MAP_SYNCMANAGER_NR(sm->sm) |
	    ESC_ADDR_MAP_SYNCMANAGER_PHYSICAL_START;
	leval16= htole16(sm->start);
	ec_datagram_write(addr, &leval16, 2);
	
	/* Length */
	addr->addr.position.off = ESC_ADDR_MAP_SYNCMANAGER_NR(sm->sm) |
	    ESC_ADDR_MAP_SYNCMANAGER_LENGTH;
	leval16 = htole16(sm->len);
	ec_datagram_write(addr, &leval16, 2);
	
	addr->addr.position.off = ESC_ADDR_MAP_SYNCMANAGER_NR(sm->sm) |
	    ESC_ADDR_MAP_SYNCMANAGER_CONTROL;
	val8 = ESC_SYNCMANAGER_CONTROL_INT_PDI;
	log_send(LOG_T_DEBUG, "PDO rx Control byte: 0x%02x", val8);
	ec_datagram_write(addr, &val8, 1);
	
	if (sm->len) {
		addr->addr.position.off = ESC_ADDR_MAP_SYNCMANAGER_NR(sm->sm) |
		    ESC_ADDR_MAP_SYNCMANAGER_ACTIVATE;
		val8 = ESC_SYNCMANAGER_ACTIVATE_ENABLE;
		ec_datagram_write(addr, &val8, 1);

		addr->addr.position.off = sm->start;

		ec_rx_pdo_add(addr, data, sm->len);
	}
	
	return 0;
}
